/*
 * bkread.c - Program to read BK001x-xx tapes
 *            bkread v1.0
 *            by Alexander (las) Lunev, 2000
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stropts.h>
#include <sys/audioio.h>
#include <string.h>
#include <signal.h>

#define MAX_SAMPLE_RATE_K 100
static int sample_rate = 44000;
static char *audio_name = AUDIO_DEVICE;
static int edge_level = 0x80;
static int pattern_count = 04200;
static int pattern_start = 1;
static int verbose = 0;
static int max_adaptations = 0;

char *last_error = NULL;
int last_pattern = 0;
           
static unsigned char data[65536];

static int period0, min_0_period, max_0_period,
           period1, min_1_period, max_1_period,
           periodS, min_S_period, max_S_period,
           max_period;

static int adapt_tries = 0;

#define PATTERN_ZERO (0)
#define PATTERN_ONE  (1)
#define PATTERN_SYNC (2)
#define PATTERN_ERR  (-1)

int audiodev;

void audioioctl(int req, audio_info_t *arg) {
    if(ioctl(audiodev, req, arg) < 0) {
        fprintf(stderr, "ioctl(audio, %d, ...) error on ", req);
        perror(audio_name);
        exit(2);
    }
}

int get_next_sample() {
    static unsigned char buff1sec[MAX_SAMPLE_RATE_K * 1000];
    static int index = MAX_SAMPLE_RATE_K * 1000;
    int j, nb;
    
    if(index >= sample_rate) {
        for(j = 0; j < sample_rate; j += nb)
          if((nb = read(audiodev, &buff1sec[j], sample_rate - j)) < 0) {
            perror(audio_name);
            exit(2);
          }
        index = 0;
    }
    return buff1sec[index++] < edge_level ? 0 : 1;
}

int read_pattern(int max) {
    static int last_sample = -1;
    int sample_count;
    int counter = 1;    

    for(; last_sample != pattern_start; last_sample = get_next_sample())
        if(max && (++counter >= max)) return -1;
    
    counter = 0;
    
    for(sample_count = 1;
        last_sample == pattern_start;
        last_sample = get_next_sample()) {
        sample_count++;
        if(max && (++counter >= max)) return -1;
    }

    return last_pattern = sample_count;
}

int find_sync() {
    int cycle_length, pattern_repeats, pattern_length;
    int tune_pattern_count = pattern_count / 021,
        sync_pattern_count = pattern_count - tune_pattern_count;
    
    if(verbose)
        fprintf(stderr, "\nListening for header synchros (S=%d; T=%d)...\n",
                sync_pattern_count, tune_pattern_count);

    for(pattern_length = -1; pattern_length < 0;) {
        for(pattern_length = pattern_repeats = 0; pattern_repeats < sync_pattern_count; ) {
            cycle_length = read_pattern(0);
            if((cycle_length >= 3) && (abs(pattern_length - cycle_length) <= 1)) {
                pattern_repeats++;
            } else {
                pattern_length = cycle_length;
                pattern_repeats = 0;
            }
        }

        if(verbose) fprintf(stderr, "Synchros detected. Pattern length = %d.\n", pattern_length);

        for(pattern_length = pattern_repeats = 0; pattern_repeats < tune_pattern_count; ) {
            if((cycle_length = read_pattern(pattern_length * 3)) >= 3) {
                pattern_length = (cycle_length + pattern_length) / 2 +
                                 (cycle_length + pattern_length) % 2;
                pattern_repeats++;
            } else { pattern_length = -1; break; }
        }

        if(verbose) {
            if(pattern_length >= 0)
                fprintf(stderr, "Pattern length after tuning = %d.\n", pattern_length);
            else
                fprintf(stderr, "Error while pattern tuning.\n");
        }
    }

    return pattern_length;
}

int analyze_pattern(int length) {
static char s[64];

    if((length >= min_0_period) && (length <= max_0_period)) {
        return PATTERN_ZERO;
    } else if((length >= min_1_period) && (length <= max_1_period)) {
        return PATTERN_ONE;
    } else if((length >= min_S_period) && (length <= max_S_period)) {
        return PATTERN_SYNC;
    } else {
        sprintf(s, "Unrecognized bit pattern (l=%d)", length);
        last_error = s;
        return PATTERN_ERR;
    }
}

void calculate_periods(int p0) {
    int i;
    
    period0 = p0;
    period1 = period0 * 2;
    periodS = period0 * 4;

    if(verbose)
        fprintf(stderr, "Patterns lengths: 0 = %d; 1 = %d.\n", period0, period1);

    max_period = period0 * 20;
    min_0_period = period0 - period0 / 2;
    max_0_period = period0 + (period1 - period0) / 2;
    min_1_period = max_0_period + 1;
    max_1_period = period1 + (periodS - period1) / 2;
    min_S_period = max_1_period + 1;
    max_S_period = periodS + (periodS - period1) / 2;        

    if(verbose) {
        fprintf(stderr, "\nSignal Period Analyzer Line (2..%d):\n", max_S_period);
        for(i = 2; i <= max_S_period; i++)
            switch(analyze_pattern(i)) {
                case PATTERN_ZERO:
                    fprintf(stderr, "0");
                    break;
                case PATTERN_ONE:
                    fprintf(stderr, "1");
                    break;
                case PATTERN_SYNC:
                    fprintf(stderr, "S");
                    break;
                case PATTERN_ERR:
                    fprintf(stderr, "-");
                    break;
                default:
                    fprintf(stderr, "?");
                    break;
            }
        fprintf(stderr, "\n");
        for(i = 2; i <= max_S_period; i++)
            fprintf(stderr, "-");
        fprintf(stderr, "\n");
        for(i = 2; i <= max_S_period; i++)
            fprintf(stderr, "%d", i / 10);
        fprintf(stderr, "\n");
        for(i = 2; i <= max_S_period; i++)
            fprintf(stderr, "%d", i % 10);
        fprintf(stderr, "\n\n");
    }
}

int try_adapt(int pattern, int to) {
    int p0, oldp0 = period0;

    if(max_adaptations && (++adapt_tries > max_adaptations)) return -1;
    
    for(p0 = oldp0 - 1; p0 <= (oldp0 + 1); p0 += 2) {
        period0 = p0;
        period1 = period0 * 2;
        periodS = period0 * 4;
        min_0_period = period0 - period0 / 2;
        max_0_period = period0 + (period1 - period0) / 2;
        min_1_period = max_0_period + 1;
        max_1_period = period1 + (periodS - period1) / 2;
        min_S_period = max_1_period + 1;
        max_S_period = periodS + (periodS - period1) / 2;
        if(analyze_pattern(pattern) == to) {
            if(verbose) fprintf(stderr, "\n*** New bit rate adaptation successfull ***\n");
            calculate_periods(p0);
            return to;
        }
    }
    period0 = oldp0;
    period1 = period0 * 2;
    periodS = period0 * 4;
    min_0_period = period0 - period0 / 2;
    max_0_period = period0 + (period1 - period0) / 2;
    min_1_period = max_0_period + 1;
    max_1_period = period1 + (periodS - period1) / 2;
    min_S_period = max_1_period + 1;
    max_S_period = periodS + (periodS - period1) / 2;
    return -1;
}   

int read_bit() {
    int prev;

    switch(analyze_pattern(read_pattern(max_period))) {
        case PATTERN_ZERO:
            prev = PATTERN_ZERO;
            break;
        case PATTERN_ONE:
            prev = PATTERN_ONE;
            break;
        case PATTERN_SYNC:
            if(try_adapt(last_pattern, PATTERN_ONE) > 0) {
                prev = PATTERN_ONE;
                break;
            }
            last_error = "PATTERN_SYNC at the start of bit";
            return -1;
        case PATTERN_ERR:
            if(try_adapt(last_pattern, PATTERN_ZERO) > 0) {
                prev = PATTERN_ZERO;
                break;
            }
            if(verbose) fprintf(stderr, "PATTERN_ERR in bit - %s.\n", last_error);
            last_error = "PATTERN_ERR at the start of bit";
            return -1;
        default:
            fprintf(stderr, "Unexpected internal error (2).\n");
            exit(2);
    }
    switch(analyze_pattern(read_pattern(max_period))) {
        case PATTERN_ZERO:
            return (prev == PATTERN_ZERO) ? 0 : 1;
        case PATTERN_ONE:
            if(try_adapt(last_pattern, PATTERN_ZERO) == PATTERN_ZERO)
                return (prev == PATTERN_ZERO) ? 0 : 1;
            fprintf(stderr, "Bit synchro = PATTERN_ONE (pl=%d, p=%d).\n", last_pattern, prev);
            last_error = "Bit synchro not PATTERN_ZERO";
            return -1;
        case PATTERN_SYNC:
            last_error = "PATTERN_SYNC at the end of bit";
            return -1;
        case PATTERN_ERR:
            if(try_adapt(last_pattern, PATTERN_ZERO) == PATTERN_ZERO)
                return (prev == PATTERN_ZERO) ? 0 : 1;
            if(verbose) fprintf(stderr, "PATTERN_ERR in bit - %s.\n", last_error);
            last_error = "PATTERN_ERR at the end of bit";
            return -1;
        default:
            fprintf(stderr, "Unexpected internal error (3).\n");
            exit(2);
    }
}

int read_byte(unsigned char *byte) {
    int i;
    
    *byte = 0;
    for(i = 0; i < 8; i++) {
        switch(read_bit()) {
            case 1:
                *byte |= 1 << i;
            case 0:
                break;
            default:
                if(verbose) fprintf(stderr, "Error reading byte - %s; bit%d; byte=%o (oct).\n", last_error, i, *byte);
                last_error = "Illegal bit pattern";
                return -1;
        }
    }
    return *byte;
}

void detect_preamble() {
    int correct_header;
    int i;
    
    for(correct_header = 0; !correct_header; ) {
        adapt_tries = 0;
        calculate_periods(find_sync());

        for(i = 1; i;) {                               // Skip sync zeros
            switch(analyze_pattern(read_pattern(max_period))) {
                case PATTERN_ZERO:
                    break;
                case PATTERN_ONE:
                    if(try_adapt(last_pattern, PATTERN_SYNC) == PATTERN_SYNC) {
                        i = 0;
                        correct_header = 1;
                        break;
                    }
                    if(try_adapt(last_pattern, PATTERN_ZERO) < 0) break;
                    if(verbose) fprintf(stderr, "Unexpected \"1\" bit synchros (pl=%d).\n", last_pattern);
                    break;
                case PATTERN_SYNC:
                    i = 0;
                    correct_header = 1;
                    break;
                case PATTERN_ERR:
                    if(try_adapt(last_pattern, PATTERN_SYNC) == PATTERN_SYNC) {
                        i = 0;
                        correct_header = 1;
                        break;
                    }
                    if(verbose) fprintf(stderr, "PATTERN_ERR in synchros - %s.\n", last_error);
                    i = 0;
                    break;
                default:
                    fprintf(stderr, "Unexpected internal error (1).\n");
                    exit(2);
            }
        }

        if(correct_header && (i = read_bit()) != 1) {
            if(verbose) {
                fprintf(stderr, "Synchro sequency doesn't finish by \"1\" bit (%d)", i);
                if(i < 0) fprintf(stderr, " - %s.\n", last_error);
                else fprintf(stderr, ".\n");
            }
            correct_header = 0;
        }
    }
}

int detect_marker() {
    int i;
    
    for(i = 0; i < 8; i++) {
        if(analyze_pattern(read_pattern(max_period)) != PATTERN_ZERO &&
           (try_adapt(last_pattern, PATTERN_ZERO) < 0)) {
            if(verbose) fprintf(stderr, "Invalid pattern in the marker synchros.\n");
            last_error = "Nonzero pattern in the marker body";
            return -1;
        }
    }
    
    if((i = analyze_pattern(read_pattern(max_period))) != PATTERN_SYNC &&
       (try_adapt(last_pattern, PATTERN_SYNC) < 0)) {
        if(i == PATTERN_ERR) {
            if(verbose) fprintf(stderr, "PATTERN_ERR detected pl=%d.\n", last_pattern);
            last_error = "Found PATTERN_ERR at the end of marker";
            return -1;
        }
        if(verbose)
            fprintf(stderr, "Marker doesn't end by PATTERN_SYNC (pl=%d).\n", last_pattern);
        last_error = "Illegal pattern at the end of marker";
        return -1;
    }

    if((i = read_bit()) != 1) {
        if(verbose) {
            fprintf(stderr, "Marker doesn't finish by \"1\" bit (%d)\n", i);
            if(i < 0) fprintf(stderr, " - %s.\n", last_error);
            else fprintf(stderr, ".\n");
        }
        last_error = "No \"1\" bit at the end of marker";
        return -1;
    }

    return 1;
}

void usage(char *arg) {
    fprintf(stderr, "Argument: %s\n", arg);
    fprintf(stderr, "Usage: bkread [-r num] [-m num] [-e num] [-z num] [-s num] [-p 0|1] [-R num] [-a num] [name]\n");
    exit(1);
}

void cancel(int sig) {
    close(audiodev);
    signal(sig, SIG_DFL);
    if(verbose) fprintf(stderr, "Signal (%d) received. Program stopped.\n", sig);
    exit(0);
}

void main(int ac, char *av[]) {
    struct stat statbuf;
    audio_info_t audio;
    int i, nb;
    int val;
    int record_gain, monitor_gain;
    int data_checksum, data_length, data_addr;
    char file_name[32];
    struct stat dummy_stat;
    unsigned int checksum;
    char *search_file = NULL;
    int dummy_read = 0;
    
    if(stat(audio_name, &statbuf) < 0) {
        fprintf(stderr, "Can't access %s\n", audio_name);
        exit(1);
    }
    if(!(statbuf.st_mode & S_IFCHR)) {
        fprintf(stderr, "%s Is not a character device\n", audio_name);
        exit(1);
    }
    if((audiodev = open(audio_name, O_RDONLY)) < 0) {
        perror(audio_name);
        exit(1);
    }
    
    signal(SIGHUP, &cancel);
    signal(SIGINT, &cancel);
    signal(SIGTERM, &cancel);
    
    AUDIO_INITINFO(&audio);
    audioioctl(AUDIO_GETINFO, &audio);
    monitor_gain = audio.monitor_gain;
    record_gain = audio.record.gain;
     
    AUDIO_INITINFO(&audio);
    audio.record.port = AUDIO_LINE_IN;
    audio.record.encoding = AUDIO_ENCODING_LINEAR;
    audio.record.sample_rate = sample_rate;
    audio.record.precision = 8;
    audio.record.channels = 1;
    audio.record.buffer_size = sample_rate * 2;
    audioioctl(AUDIO_SETINFO, &audio);
    
    for(i = 1; i < ac; i++) {
        if(*av[i] == '-') {
	if(strlen(av[i]) != 2) usage(av[i]);
          switch(*(av[i] + 1)) {
            case 'r':
                if(++i >= ac) usage("-r without value");
                val = atoi(av[i]);
                if((val < 0) || (val > 100))
                  usage("-r num - out of range, must be in range 0..100");
                AUDIO_INITINFO(&audio);
                record_gain = audio.record.gain = (AUDIO_MAX_GAIN - AUDIO_MIN_GAIN) * val / 100 + AUDIO_MIN_GAIN;
                audioioctl(AUDIO_SETINFO, &audio);
                break;
            case 'm':
                if(++i >= ac) usage("-m without value");
                val = atoi(av[i]);
                if((val < 0) || (val > 100))
                  usage("-m: num out of range, must be in range 0..100");
                AUDIO_INITINFO(&audio);
                monitor_gain = audio.monitor_gain = (AUDIO_MAX_GAIN - AUDIO_MIN_GAIN) * val / 100 + AUDIO_MIN_GAIN;
                audioioctl(AUDIO_SETINFO, &audio);
                break;
            case 'e':
                if(++i >= ac) usage("-e without value");
                val = atoi(av[i]);
                if((val < 0) || (val > 255))
                  usage("-e: num out of range, must be in range 0..255");
                edge_level = val;
                break;
            case 's':
                if(++i >= ac) usage("-s without value");
                val = atoi(av[i]);
                if(val <= 10)
                  usage("-s: num less than ten");
                pattern_count = val;
                break;
            case 'p':
                if(++i >= ac) usage("-p without value");
                val = atoi(av[i]);
                if((val < 0) || (val > 1))
                  usage("-s: num must be 0 or 1");
                pattern_start = val;
                break;
            case 'R':
                if(++i >= ac) usage("-R without value");
                val = atoi(av[i]);
                if((val < 8) || (val > MAX_SAMPLE_RATE_K))
                  usage("-R: num must be between 8 or 100 (K)"); // MAX_SAMPLE_RATE_K
                AUDIO_INITINFO(&audio);
                sample_rate = audio.record.sample_rate = val * 1000;
                audioioctl(AUDIO_SETINFO, &audio);
                break;
            case 'a':
                if(++i >= ac) usage("-a without value");
                val = atoi(av[i]);
                max_adaptations = val;
                break;
            case 'v':
                verbose = 1;
                break;
            default:
                usage(av[i]);
          }
	} else {
	    if(search_file != NULL) usage(av[i]);
	    search_file = av[i];
	}
    }
    
    if(verbose) {
        fprintf(stderr, "Audio:    recording gain = %d\n", record_gain);
        fprintf(stderr, "Audio:    monitor gain = %d\n", monitor_gain);
        fprintf(stderr, "Audio:    sample_rate = %d\n", sample_rate);
        fprintf(stderr, "Analyzer: edge level = %d\n", edge_level);
        fprintf(stderr, "Reader:   Synchro tuning pattern count = %d\n", pattern_count);
        if(max_adaptations) fprintf(stderr, "Reader:   Max. number of adaptations = %d\n", max_adaptations);
    }
//
// Waiting for a file header synchros
//
    for(;;) {
        detect_preamble();
        if(detect_marker() < 0) {
            if(verbose)
                fprintf(stderr, "Can't detect marker: %s.\n", last_error);
            continue;
        }
    
        for(i = 0; i < 20; i++)
            if(read_byte(&data[i]) < 0) {
                if(verbose) fprintf(stderr, "Error in header (%s); offset = %o (octal).\n", last_error, i);
                break;
            }
        if(i != 20) continue;

        data_addr = data[0] + (data[1] << 8);
        data_length = data[2] + (data[3] << 8);
        memcpy(file_name, &data[4], 16);
        file_name[16] = '\0';
        for(i = 15; i >= 0; i--)
            if(file_name[i] != ' ') break;
            else file_name[i] = '\0';
    
        if(verbose) {
            fprintf(stderr, "File Name = \"%-16s\"\n", file_name);
            fprintf(stderr, "Data Length = %6o (octal).\n", data_length);
            fprintf(stderr, "Data Address = %6o (octal).\n", data_addr);
        } else fprintf(stderr, "%s\n", file_name);
	
        if(search_file != NULL) {
	    if(strcmp(search_file, file_name)) dummy_read = 1; else dummy_read = 0;
	}
    
        if((val = read_bit()) != 0 ||
           (val = read_bit()) != 0 ||
           (val = read_bit()) != 0 ||
           (val = read_bit()) != 0) {
            if(verbose)
                fprintf(stderr, "Can't read finishing \"0000\" at the end of header (%d).\n", val);
            continue;
        }
        
        if((i = analyze_pattern(read_pattern(max_period))) != PATTERN_SYNC) {
            if(verbose) {
                fprintf(stderr, "Can't find SYNC at the end of header (%d)", i);
                if(i < 0) fprintf(stderr, " - %s.\n", last_error);
                else fprintf(stderr, ".\n");
            }
            continue;
        }
        
        if((i = read_bit()) != 1) {
            if(verbose) {
                fprintf(stderr, "Header doesn't finish by \"1\" bit (%d)", i);
                if(i < 0) fprintf(stderr, " - %s.\n", last_error);
                else fprintf(stderr, ".\n");
            }
            continue;
        }

        for(nb = 20; nb < data_length + 22; nb++) {
            if(read_byte(&data[nb]) < 0) {
                if(nb >= (data_length + 20)) {
                    fprintf(stderr, "Error reading checksum byte %6o (octal): %s.\n", nb - 20 - data_length, last_error);
                    nb = data_length + 22;
                    data[nb - 2] = data[nb - 1] = 0;
                } else {
                    fprintf(stderr, "Error reading data byte %6o (octal): %s.\n", nb - 20, last_error);
                }
                
                break;
            }
        }
        if(nb != data_length + 22) continue;

        data_checksum = data[nb - 2] + (data[nb - 1] << 8);
        if(verbose) {
            fprintf(stderr, "Tape Checksum = %6o (octal).\n", data_checksum);
        }
        
        for(checksum = 0, i = 20; i < nb - 2; i++) {
            checksum += data[i];
            checksum += (checksum & 0200000) != 0;
            checksum &= 0177777;
        }
        
        if(verbose) {
            fprintf(stderr, "Data Checksum = %6o (octal).\n", checksum);
        }
        if(checksum != data_checksum) {
            fprintf(stderr, "Warning: checksum error.\n");
        }
        
        if(!dummy_read) {
    	    val = strlen(file_name);
    	    for(i = 1; stat(file_name, &dummy_stat) >= 0; i++) {
        	sprintf(&file_name[val], ".%d", i);
    	    }
        
    	    if(verbose) fprintf(stderr, "Saving file to: %s (%d bytes).\n", file_name, nb);
        
    	    if(((i = open(file_name, O_WRONLY | O_CREAT, 0644)) < 0) ||
               (write(i, data, nb) != nb)) {
        	fprintf(stderr, "Can't save file %s; ", file_name);
        	perror("open()|write()");
    	    } else if(!verbose) fprintf(stderr, "%d bytes written.\n", nb);
    	    close(i);
	}
    }
}
